@@ -1118,7 +1118,7 @@ class Matrix
   #     => 67 96
   #        48 99
   #
-  def ** (other)
+  def **(other)
     case other
     when Integer
       x = self
@@ -1644,7 +1644,7 @@ class Matrix
       end
     end
 
-    def / (other)
+    def /(other)
       case other
       when Numeric
         Scalar.new(@value / other)
@@ -1657,7 +1657,7 @@ class Matrix
       end
     end
 
-    def ** (other)
+    def **(other)
       case other
       when Numeric
         Scalar.new(@value ** other)
@@ -1685,6 +1685,7 @@ end
 # * Vector.[](*array)
 # * Vector.elements(array, copy = true)
 # * Vector.basis(size: n, index: k)
+# * Vector.zero(n)
 #
 # To access elements:
 # * #[](i)
@@ -1697,6 +1698,7 @@ end
 # * #angle_with(v)
 # * Vector.independent?(*vs)
 # * #independent?(*vs)
+# * #zero?
 #
 # Vector arithmetic:
 # * #*(x) "is matrix or number"
@@ -1768,6 +1770,17 @@ class Vector
     new convert_to_array(array, false)
   end
 
+  #
+  # Return a zero vector.
+  #
+  #    Vector.zero(3) => Vector[0, 0, 0]
+  #
+  def Vector.zero(size)
+    raise ArgumentError, "invalid size (#{size} for 0..)" if size < 0
+    array = Array.new(size, 0)
+    new convert_to_array(array, false)
+  end
+
   #
   # Vector.new is private; use Vector[] or Vector.elements to create.
   #
@@ -1882,6 +1895,13 @@ class Vector
     self.class.independent?(self, *vs)
   end
 
+  #
+  # Returns +true+ iff all elements are zero.
+  #
+  def zero?
+    all?(&:zero?)
+  end
+
   #--
   # COMPARING -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
   #++
