@@ -3,7 +3,7 @@
 
   thread_pthread.c -
 
-  $Author: nobu $
+  $Author: normal $
 
   Copyright (C) 2004-2007 Koichi Sasada
 
@@ -740,6 +740,12 @@ ruby_init_stack(volatile VALUE *addr
     )
 {
     native_main_thread.id = pthread_self();
+#ifdef __ia64
+    if (!native_main_thread.register_stack_start ||
+        (VALUE*)bsp < native_main_thread.register_stack_start) {
+        native_main_thread.register_stack_start = (VALUE*)bsp;
+    }
+#endif
 #if MAINSTACKADDR_AVAILABLE
     if (native_main_thread.stack_maxsize) return;
     {
@@ -749,7 +755,7 @@ ruby_init_stack(volatile VALUE *addr
 	    native_main_thread.stack_maxsize = size;
 	    native_main_thread.stack_start = stackaddr;
 	    reserve_stack(stackaddr, size);
-	    return;
+	    goto bound_check;
 	}
     }
 #endif
@@ -762,12 +768,6 @@ ruby_init_stack(volatile VALUE *addr
                     native_main_thread.stack_start < addr)) {
         native_main_thread.stack_start = (VALUE *)addr;
     }
-#endif
-#ifdef __ia64
-    if (!native_main_thread.register_stack_start ||
-        (VALUE*)bsp < native_main_thread.register_stack_start) {
-        native_main_thread.register_stack_start = (VALUE*)bsp;
-    }
 #endif
     {
 #if defined(HAVE_GETRLIMIT)
@@ -797,6 +797,9 @@ ruby_init_stack(volatile VALUE *addr
 #endif
     }
 
+#if MAINSTACKADDR_AVAILABLE
+  bound_check:
+#endif
     /* If addr is out of range of main-thread stack range estimation,  */
     /* it should be on co-routine (alternative stack). [Feature #2294] */
     {
@@ -829,8 +832,8 @@ native_thread_init_stack(rb_thread_t *th)
     rb_nativethread_id_t curr = pthread_self();
 
     if (pthread_equal(curr, native_main_thread.id)) {
-	th->machine.stack_start = native_main_thread.stack_start;
-	th->machine.stack_maxsize = native_main_thread.stack_maxsize;
+	th->ec.machine.stack_start = native_main_thread.stack_start;
+	th->ec.machine.stack_maxsize = native_main_thread.stack_maxsize;
     }
     else {
 #ifdef STACKADDR_AVAILABLE
@@ -838,11 +841,11 @@ native_thread_init_stack(rb_thread_t *th)
 	size_t size;
 
 	if (get_stack(&start, &size) == 0) {
-	    th->machine.stack_start = start;
-	    th->machine.stack_maxsize = size;
+	    th->ec.machine.stack_start = start;
+	    th->ec.machine.stack_maxsize = size;
 	}
 #elif defined get_stack_of
-	if (!th->machine.stack_maxsize) {
+	if (!th->ec.machine.stack_maxsize) {
 	    native_mutex_lock(&th->interrupt_lock);
 	    native_mutex_unlock(&th->interrupt_lock);
 	}
@@ -851,9 +854,9 @@ native_thread_init_stack(rb_thread_t *th)
 #endif
     }
 #ifdef __ia64
-    th->machine.register_stack_start = native_main_thread.register_stack_start;
-    th->machine.stack_maxsize /= 2;
-    th->machine.register_stack_maxsize = th->machine.stack_maxsize;
+    th->ec.machine.register_stack_start = native_main_thread.register_stack_start;
+    th->ec.machine.stack_maxsize /= 2;
+    th->ec.machine.register_stack_maxsize = th->ec.machine.stack_maxsize;
 #endif
     return 0;
 }
@@ -881,7 +884,7 @@ thread_start_func_1(void *th_ptr)
 	native_thread_init(th);
 	/* run */
 #if defined USE_NATIVE_THREAD_INIT
-	thread_start_func_2(th, th->machine.stack_start, rb_ia64_bsp());
+	thread_start_func_2(th, th->ec.machine.stack_start, rb_ia64_bsp());
 #else
 	thread_start_func_2(th, &stack_start, rb_ia64_bsp());
 #endif
@@ -1003,10 +1006,10 @@ native_thread_create(rb_thread_t *th)
 	const size_t stack_size = th->vm->default_params.thread_machine_stack_size;
 	const size_t space = space_size(stack_size);
 
-        th->machine.stack_maxsize = stack_size - space;
+        th->ec.machine.stack_maxsize = stack_size - space;
 #ifdef __ia64
-        th->machine.stack_maxsize /= 2;
-        th->machine.register_stack_maxsize = th->machine.stack_maxsize;
+        th->ec.machine.stack_maxsize /= 2;
+        th->ec.machine.register_stack_maxsize = th->ec.machine.stack_maxsize;
 #endif
 
 #ifdef HAVE_PTHREAD_ATTR_INIT
@@ -1029,8 +1032,8 @@ native_thread_create(rb_thread_t *th)
 #ifdef get_stack_of
 	if (!err) {
 	    get_stack_of(th->thread_id,
-			 &th->machine.stack_start,
-			 &th->machine.stack_maxsize);
+			 &th->ec.machine.stack_start,
+			 &th->ec.machine.stack_maxsize);
 	}
 	native_mutex_unlock(&th->interrupt_lock);
 #endif
@@ -1315,17 +1318,21 @@ void
 rb_thread_wakeup_timer_thread(void)
 {
     /* must be safe inside sighandler, so no mutex */
-    ATOMIC_INC(timer_thread_pipe.writing);
-    rb_thread_wakeup_timer_thread_fd(&timer_thread_pipe.normal[1]);
-    ATOMIC_DEC(timer_thread_pipe.writing);
+    if (timer_thread_pipe.owner_process == getpid()) {
+	ATOMIC_INC(timer_thread_pipe.writing);
+	rb_thread_wakeup_timer_thread_fd(&timer_thread_pipe.normal[1]);
+	ATOMIC_DEC(timer_thread_pipe.writing);
+    }
 }
 
 static void
 rb_thread_wakeup_timer_thread_low(void)
 {
-    ATOMIC_INC(timer_thread_pipe.writing);
-    rb_thread_wakeup_timer_thread_fd(&timer_thread_pipe.low[1]);
-    ATOMIC_DEC(timer_thread_pipe.writing);
+    if (timer_thread_pipe.owner_process == getpid()) {
+	ATOMIC_INC(timer_thread_pipe.writing);
+	rb_thread_wakeup_timer_thread_fd(&timer_thread_pipe.low[1]);
+	ATOMIC_DEC(timer_thread_pipe.writing);
+    }
 }
 
 /* VM-dependent API is not available for this function */
@@ -1512,7 +1519,7 @@ native_set_thread_name(rb_thread_t *th)
 	    SET_CURRENT_THREAD_NAME(RSTRING_PTR(loc));
 	}
 	else if (!NIL_P(loc = rb_proc_location(th->first_proc))) {
-	    const VALUE *ptr = RARRAY_CONST_PTR(loc); /* [ String, Fixnum ] */
+	    const VALUE *ptr = RARRAY_CONST_PTR(loc); /* [ String, Integer ] */
 	    char *name, *p;
 	    char buf[16];
 	    size_t len;
@@ -1537,6 +1544,17 @@ native_set_thread_name(rb_thread_t *th)
 #endif
 }
 
+static VALUE
+native_set_another_thread_name(rb_nativethread_id_t thread_id, VALUE name)
+{
+#ifdef SET_ANOTHER_THREAD_NAME
+    const char *s = "";
+    if (!NIL_P(name)) s = RSTRING_PTR(name);
+    SET_ANOTHER_THREAD_NAME(thread_id, s);
+#endif
+    return name;
+}
+
 static void *
 thread_timer(void *p)
 {
@@ -1584,6 +1602,7 @@ rb_thread_create_timer_thread(void)
 	int err;
 #ifdef HAVE_PTHREAD_ATTR_INIT
 	pthread_attr_t attr;
+	rb_vm_t *vm = GET_VM();
 
 	err = pthread_attr_init(&attr);
 	if (err != 0) {
@@ -1598,9 +1617,17 @@ rb_thread_create_timer_thread(void)
 	     * at least 16KB (4 pages).  FreeBSD 8.2 AMD64 causes
 	     * machine stack overflow only with PTHREAD_STACK_MIN.
 	     */
+	    enum {
+		needs_more_stack =
+#if defined HAVE_VALGRIND_MEMCHECK_H && defined __APPLE__
+		1
+#else
+		THREAD_DEBUG != 0
+#endif
+	    };
 	    size_t stack_size = PTHREAD_STACK_MIN; /* may be dynamic, get only once */
 	    if (stack_size < min_size) stack_size = min_size;
-	    if (THREAD_DEBUG) stack_size += BUFSIZ;
+	    if (needs_more_stack) stack_size += BUFSIZ;
 	    pthread_attr_setstacksize(&attr, stack_size);
 	}
 # endif
@@ -1620,10 +1647,20 @@ rb_thread_create_timer_thread(void)
 	    rb_bug("rb_thread_create_timer_thread: Timer thread was already created\n");
 	}
 #ifdef HAVE_PTHREAD_ATTR_INIT
-	err = pthread_create(&timer_thread.id, &attr, thread_timer, &GET_VM()->gvl);
+	err = pthread_create(&timer_thread.id, &attr, thread_timer, &vm->gvl);
 	pthread_attr_destroy(&attr);
+
+	if (err == EINVAL) {
+	    /*
+	     * Even if we are careful with our own stack use in thread_timer(),
+	     * any third-party libraries (eg libkqueue) which rely on __thread
+	     * storage can cause small stack sizes to fail.  So lets hope the
+	     * default stack size is enough for them:
+	     */
+	    err = pthread_create(&timer_thread.id, NULL, thread_timer, &vm->gvl);
+	}
 #else
-	err = pthread_create(&timer_thread.id, NULL, thread_timer, &GET_VM()->gvl);
+	err = pthread_create(&timer_thread.id, NULL, thread_timer, &vm->gvl);
 #endif
 	if (err != 0) {
 	    rb_warn("pthread_create failed for timer: %s, scheduling broken",
@@ -1712,8 +1749,8 @@ ruby_stack_overflowed_p(const rb_thread_t *th, const void *addr)
     else
 #endif
     if (th) {
-	size = th->machine.stack_maxsize;
-	base = (char *)th->machine.stack_start - STACK_DIR_UPPER(0, size);
+	size = th->ec.machine.stack_maxsize;
+	base = (char *)th->ec.machine.stack_start - STACK_DIR_UPPER(0, size);
     }
     else {
 	return 0;
